from distutils.core import setup, Extension
from Cython.Build import cythonize
import numpy 

extensions = [Extension(
        name="solve",
        sources=["solve.pyx"],
        include_dirs=[numpy.get_include()],
        )
]

setup(
    ext_modules = cythonize(extensions),
)



from optoanalysis import DataObject
from optosim.sde_solver import sde_solver
import numpy as _np
from multiprocessing import Pool as _Pool
from frange import frange

class SimData(DataObject):
    """
    Simualtes a particle by solving the SDE for a single degree of freedom for each 
    trap frequency value in TrapFreqArray (and values of eta [modulation depth] in 
    etaArray and using the values specified by the other parameters to this object 
    and then convolving the signals from each by addition in time-space along with 
    addition of white noise produced by numpy.random.normal with a std deviation
    given by NoiseStdDev. 

    Attributes
    ----------
    q0 : float
        
    v0 : float
        
    TimeTuple : tuple
        
    timeStart : float
        
    timeEnd : float
        
    SampleFreq : float
        
    TrapFreqArray : ndarray
        
    Gamma0 : float
        Enviromental damping
    mass : float
        mass of nanoparticle (in Kg)
    ConvFactor : float
        
    NoiseStdDev : float
        
    T0 : float, optional
        
    deltaGammaArray : ndarray
        array containing the additional damping on each degree of freedom 
        due to other effects (e.g. feedback cooling)

    dt : float
        
    seed : float
        random seed for generating the weiner paths for SDE solving
        defaults to None i.e. no seeding of random numbers
        sets the seed prior to initialising the SDE solvers such
        that the data is repeatable but that each solver uses
        different random numbers
    dtSample : float
        




    DownSampleAmount : int
        
    timeStep : float
        
    time : ndarray
        
    simtime : frange
        
    Noise : ndarray
        
    voltage : ndarray
        
    TrueSignals : ndarray
        



    Omega0 : float
        Trapping frequency
    Gamma0 : float
        Enviromental damping
    mass : float

    eta : float, optional
        modulation depth (as a fraction), defaults to 0
    T0 : float, optional
        Temperature of the environment, defaults to 300

    q0 : float, optional
        initial position, defaults to 0
    v0 : float, optional
        intial velocity, defaults to 0

    dt : float, optional
        time interval for simulation / solver
    seed : float, optional
        random seed for generate_weiner_path, defaults to None
        i.e. no seeding of random numbers

    """
    def __init__(self, TimeTuple, SampleFreq, TrapFreqArray, Gamma0, mass, ConvFactorArray, NoiseStdDev, T0=300.0, deltaGammaArray=None, dt=1e-9, seed=None, NPerSegmentPSD=1000000):
        """
        Parameters
        ----------
        TimeTuple : tuple, optional
            tuple of start and stop time for simulation / solver
        SampleFreq : float
            Sample freq to downsample data to (Hz)
        TrapFreqArray : ndarray
            Array of trap frequencies of Z, X and Y motion (Hz).
        Gamma0 : float
            Damping due to the enviroment (radians/second)
        mass : float
            mass of the nanoparticle (kg)
        ConvFactorArray : ndarray
            Conversion factors to use to go from motion in nms
            to signal in volts for each degree of freedom
        NoiseStdDev : float
            std deviation of white noise applied to particle signal
            after generation from SDE solving.
        T0 : float, optional
            Temperature of the environment
        deltaGammaArray : ndarray, optional
            array containing the additional damping on each degree of freedom 
            due to other effects (e.g. feedback cooling)
        dt : float, optional
            time step for SDE solver
        seed : float, optional
            random seed for generating the weiner paths for SDE solving
            defaults to None i.e. no seeding of random numbers
            sets the seed prior to initialising the SDE solvers such
            that the data is repeatable but that each solver uses
            different random numbers. No seed by default.
        NPerSegnmentPSD : int, optional
            number of points per segment to use in calculating the PSD


        """
        self.q0 = 0.0
        self.v0 = 0.0
        self.TimeTuple = (TimeTuple[0], TimeTuple[1])
        self.timeStart = TimeTuple[0]
        self.timeEnd = TimeTuple[1]
        self.SampleFreq = SampleFreq
        self.TrapFreqArray = _np.array(TrapFreqArray)
        self.Gamma0 = Gamma0
        if deltaGammaArray == None:
            self.deltaGammaArray = _np.zeros_like(TrapFreqArray)
        else:
            if len(deltaGammaArray) != len(TrapFreqArray):
                raise("deltaGammaArray should be the same length as TrapFreqArray")
            self.deltaGammaArray = deltaGammaArray
        self.mass = mass
        self.ConvFactorArray = ConvFactorArray
        self.NoiseStdDev = NoiseStdDev
        self.T0 = T0
        self.dt = dt
        self.seed = seed        
        dtSample = 1/SampleFreq
        self.DownSampleAmount = round(dtSample/dt)
        self.timeStep = dtSample/dt
        if _np.isclose(dtSample/dt, self.DownSampleAmount, atol=1e-6) == False:
            raise ValueError("The sample rate {} has a time interval between samples of {}, this is not a multiple of the simualted time interval {}. dtSample/dt = {}".format(SampleFreq, dtSample, dt, self.timeStep))
        self.generate_simulated_data() # solves SDE for each frequency and eta value specified
        # along requested time interval
        self.time = frange(TimeTuple[0], TimeTuple[1], self.DownSampleAmount*dt)
        self.simtime = frange(TimeTuple[0], TimeTuple[1], dt)
        self.Noise = _np.random.normal(0, self.NoiseStdDev, len(self.time))
        self.voltage = _np.copy(self.Noise)
        self.TrueSignals = []
        for i, sdesolver in enumerate(self.sde_solvers):
            self.TrueSignals.append(_np.array([sdesolver.q, sdesolver.v]))
            self.voltage += ConvFactorArray[i]*sdesolver.q[::self.DownSampleAmount]
        self.TrueSignals = _np.array(self.TrueSignals)
        self.get_PSD(NPerSegmentPSD)
        del(self.sde_solvers)
        return None

    def generate_simulated_data(self):
        self.sde_solvers = []
        if self.seed != None:
            _np.random.seed(self.seed)
        for i, freq in enumerate(self.TrapFreqArray):
            TrapOmega = freq*2*_np.pi
            solver = sde_solver(TrapOmega, self.Gamma0, self.deltaGammaArray[i], self.mass, T0=self.T0, q0=self.q0, v0=self.v0, TimeTuple=self.TimeTuple, dt=self.dt)
            self.sde_solvers.append(solver)
        #workerPool = _Pool()
        #workerPool.map(run_solve, self.sde_solvers)
        for solver in self.sde_solvers:
            print('solving...')
            solver.solve()
        return None

def run_solve(sde_solver):
    print('solving...')
    sde_solver.q, sde_solver.v = sde_solver.solve()
    return None



cimport numpy as np
cimport cython

@cython.boundscheck(False) # Turns off IndexError type warnings - e.g. a = [1, 2, 3]; a[5]
@cython.wraparound(False) # Turns off Python a[-1] indexing - will segfault.
@cython.overflowcheck(False) # Check that no integer overflows occur with arithmetic operations
@cython.initializedcheck(False) # Checks that memory is initialised when passed in
cpdef solve(np.ndarray[double, ndim=1] q,
            np.ndarray[double, ndim=1] v,
            double dt,
            np.ndarray[double, ndim=1] dwArray,
            double Gamma0,
            double deltaGamma,
            double Omega0,
            double b_v,
            np.ndarray[double, ndim=1] SqueezingPulseArray,
            int startIndex,
            int NumTimeSteps ):
    """
    Solves the SDE specified in sde_solver.py using cythonized python code.
    
    Parameters
    ----------
    q : ndarray
        intialised array of positions (with q[0] = intial position)
    v : ndarray
        intialised array of velocities (with v[0] = intial velocity)
    dt : float
        time interval / step for simulation / solver on which to solve SDE
    dwArray : ndarray
        random values to use for Weiner process
    Gamma0 : float
        Enviromental damping parameter (angular frequency - radians/s)
    deltaGamma : float
        damping due to other effects (e.g. feedback cooling) (radians/s)
    Omega0 : float
        Trapping frequency (angular frequency - radians/s)
    eta : float
        modulation depth (as a fraction)
    b_v : float
        term multiplying the Weiner process in the SDE sqrt(2*Γ0*kB*T0/m)
    SqueezingPulseArray : ndarray
        Array of values containing modulation depth of squeezing pulses (as a decimal i.e. 1 = 100%)
    startIndex : int
        array index (of q and v) at which to start solving the SDE
    NumTimeSteps : int
        The length of the time array minus 1 (number of time points over which
        to solve the SDE)

    Returns
    -------
    q : ndarray
        array of positions with time found from solving the SDE
    v : ndarray
        array of velocities with time found from solving the SDE
    """
    cdef int n
    for n in range(startIndex, NumTimeSteps+startIndex):
        v[n+1] = v[n] + (-(Gamma0 + deltaGamma)*v[n] - SqueezingPulseArray[n]*Omega0**2*q[n])*dt + b_v*dwArray[n]
        q[n+1] = q[n] + v[n]*dt
    return q, v





from sys import version
print(version)
if version[0] != '3':
    raise OSError("This package requires python 3")
from setuptools import setup
from setuptools.extension import Extension
import pip
pip.main(['install', '-r', 'requirements.txt']) # installs requirements from requirements.txt file
import os
try:
    import numpy
except ModuleNotFoundError:
    raise Exception("Please install numpy and attempt re-installing. Try 'pip install numpy'")
try:
    from Cython.Build import cythonize
    from Cython.Build import build_ext
except ModuleNotFoundError:
    raise Exception("Please install Cython and attempt re-installing. Try 'pip install Cython'")

    
mypackage_root_dir = os.path.dirname(__file__)
with open(os.path.join(mypackage_root_dir, 'requirements.txt')) as requirements_file:
    requirements = requirements_file.read().splitlines()

with open(os.path.join(mypackage_root_dir, 'optosim/VERSION')) as version_file:
    version = version_file.read().strip()

extensions = [Extension(
    name="solve",
    sources=["optosim/sde_solver/solve.pyx"],
    include_dirs=[numpy.get_include()],
    )
]

setup(name='optosim',
      version=version,
      description='Python package with functions for data analysis',
      author='Ashley Setter',
      author_email='A.Setter@soton.ac.uk',
      url="https://github.com/AshleySetter/optosim",
      download_url="https://github.com/AshleySetter/optosim/archive/{}.tar.gz".format(version),
      include_package_data=True,
      packages=['optosim',
                'optosim.sde_solver',
      ],
      ext_modules = cythonize(extensions),
      install_requires=requirements,
)



from scipy.constants import Boltzmann
import numpy as np
import os
if 'READTHEDOCS' not in os.environ:
    from solve import solve as solve_cython
from frange import frange

class sde_solver():
    """
    Solves the following SDE for q(t) and d(q(t))/d(t):
    d^2(q(t))/d(t)^2 = -[Γ0 - Ω0 η q(t)^2]*d(q(t))/d(t) - Ω0^2*q(t)^2 + sqrt(2*Γ0*kB*T0/m)*d(W(t))/d(t)
    
    Using the Euler-Maruyama method.

    Where:
    q(t) is the position of the particle (x, y or z) with time
    Ω0 is the trapping frequency
    Γ0 is the damping due to the environment
    kB is the Boltzmann constant
    T0 is the environment temperature
    m is the mass of the nanoparticle
    η is the modulation depth of the cooling signal ???
    W(t) is the Wiener process
    """
    def __init__(self, Omega0, Gamma0, deltaGamma, mass, T0=300, q0=0, v0=0, TimeTuple=[0, 100e-6], dt=1e-9, seed=None):
        """
        Initialises the sde_solver instance.

        Parameters
        ----------
        Omega0 : float
            Trapping frequency
        Gamma0 : float
            Enviromental damping
        deltaGamma : float
            damping due to other effects (e.g. feedback cooling)
        mass : float
            mass of nanoparticle (in Kg)
        T0 : float, optional
            Temperature of the environment, defaults to 300
        q0 : float, optional
            initial position, defaults to 0
        v0 : float, optional
            intial velocity, defaults to 0
        TimeTuple : tuple, optional
            tuple of start and stop time for simulation / solver
        dt : float, optional
            time interval for simulation / solver
        seed : float, optional
            random seed for generate_weiner_path, defaults to None
            i.e. no seeding of random numbers
        
        """
        self.k_B = Boltzmann # J/K
        self.tArray = np.arange(0, 500e-6, dt)
        self.q0 = q0
        self.v0 = v0
        self.Omega0 = Omega0
        self.Gamma0 = Gamma0
        self.deltaGamma = deltaGamma
        self.mass = mass
        self.T0 = T0
        self.TimeTuple = TimeTuple
        self.b_v = np.sqrt(2*self.Gamma0*self.k_B*self.T0/self.mass) # a constant
        self.dt = dt
        self.tArray = frange(TimeTuple[0], TimeTuple[1], dt)
        self.generate_weiner_path(seed)
        
        self.q = np.zeros(len(self.tArray)) # initialises position array, q
        self.v = np.zeros(len(self.tArray)) # initialises velocity array, v
        self.q[0] = self.q0 # sets initial position to q0
        self.v[0] = self.v0 # sets initial position to v0
        self.SqueezingPulseArray = np.ones(len(self.tArray)) # initialises squeezing pulse array such that there is no squeezing
        return None

    def generate_weiner_path(self, seed=None):
        """
        Generates random values of dW along path and populates
        self.dwArray property with values along with path.
        The values of dW are independent and identically 
        distributed normal random variables with expected 
        value 0 and variance dt. Sets the dwArray property.

        seed : float, optional
            random seed for generate_weiner_path, defaults to None
            i.e. no seeding of random numbers

        """
        if seed != None:
            np.random.seed(seed)
        self.dwArray = np.random.normal(0, np.sqrt(self.dt), len(self.tArray))
        return None

#    def a_q(self, t, p, q): # replaced with just p to reduce slow python function evaluations
#        return p

    def a_v(self, q, v):
        return _a_v(q, v, self.Gamma0, self.Omega0, self.eta)
        
    def solve(self, NumTimeSteps=None, startIndex=0):
        """
        Solves the SDE from timeTuple[0] to timeTuple[1]
        
        Parameters
        ----------
        NumTimeSteps : int, optional
            number of time steps to solve for
        startIndex : int, optional
            array index (of q and v) at which to start solving the SDE

        Returns
        -------
        self.q : ndarray
            array of positions with time
        self.v : ndarray
            array of velocities with time
        """
        if NumTimeSteps == None:
            NumTimeSteps = (len(self.tArray) - 1) - startIndex
        self.q, self.v = solve_cython(self.q, self.v, float(self.dt), self.dwArray, float(self.Gamma0),float(self.deltaGamma), float(self.Omega0), float(self.b_v), SqueezingPulseArray=self.SqueezingPulseArray, startIndex=startIndex, NumTimeSteps=NumTimeSteps)
        return self.q, self.v

#def _a_v(q, v, Gamma0, Omega0, eta):
#    return -(Gamma0 + deltaGamma)*v - Omega0**2*q

